

import commandcenter.CommandCenter;
import enumerate.Action;
import gameInterface.AIInterface;
import myai.GarnetNormalAI;
import myai.GarnetSpeedAI;
import myai.LudNormalAI;
import myai.LudSpeedAI;
import myai.MyAI;
import myai.ZenNormalAI;
import myai.ZenSpeedAI;
import myenumurate.CharaName;
import myenumurate.GameMode;
import structs.FrameData;
import structs.GameData;
import structs.Key;




/**
 * MinMaxとMCTSにより実装するAI
 * サンプルのMCTSにはムダな行動が多いため、
 * 以下のデメリットがある
 * ・探索空間が広すぎて計算時間を有効利用できない
 * ・あたらない攻撃をしてスキができたところに攻撃される
 * ・あたらない攻撃をしてエネルギーの無駄遣いをする
 * これらのデメリットを避けるため、ムダな行動については枝刈りをする
 *
 * 自分も敵も何もしないまましばらく待った状態をNonFrameDataとする。
 * 敵が何もしない前提で味方があるActionをした場合、NonFrameDataよりいい状態になったのであれば、
 * そのActionはやる価値がある。
 * これをHittingActionと呼ぶ。（敵に攻撃があたる、もしくは敵の攻撃を避けることができる）
 * MCTSの初動をHittingActionの中から選択する。
 * 注意点として、サンプルのNodeの実装だと、
 * 初動を枝刈りすると2手先以降もすべて枝刈りされた状態になり、想定しているMCTSができなくなる。
 * そのため、初動のみ枝刈りできるようにNodeも一部実装を変更している。
 *
 * @author Eita Aoki
 */
public class GigaThunder implements AIInterface {

	private Key inputKey;
	private boolean player;
	private FrameData frameData;
	private CommandCenter cc;
	private GameData gd;


	private boolean firstflag=true;
	private GameMode gamemode;
	private CharaName charname;

	MyAI ai;

	@Override
	public void close() {
		// Nothing to do here
	}

	@Override
	public String getCharacter() {
		// この関数明らかにいらなくないですか？3年か4年ぐらい前のルールの名残ですよね？
		return CHARACTER_ZEN;
	}

	/*
	 * 初めてgetInformationが呼ばれた際にキャラやゲームモードごとのAIを呼び出す
	 *
	 */
	public void firstgetInformation(){
		System.out.println("call firstgetInformation!");
		CommandCenter c=new CommandCenter();c.setFrameData(frameData, player);

		int myhp=c.getMyHP();
		int ophp=c.getEnemyHP();
		if(myhp<1000&&ophp<1000)this.gamemode=GameMode.NORMAL;
		else if(myhp>1000&&ophp<1000)this.gamemode=GameMode.SPEED;
		else this.gamemode=GameMode.LUDTRAIN;

		if(this.charname==CharaName.ZEN&&this.gamemode==GameMode.NORMAL)
			ai=new ZenNormalAI(gd,player);
		else if(this.charname==CharaName.ZEN&&this.gamemode==GameMode.SPEED)
			ai=new ZenSpeedAI(gd,player);
		else  if(this.charname==CharaName.GARNET&&this.gamemode==GameMode.NORMAL)
			ai=new GarnetNormalAI(gd,player);
		else  if(this.charname==CharaName.GARNET&&this.gamemode==GameMode.SPEED)
			ai=new GarnetSpeedAI(gd,player);
		else  if(this.charname==CharaName.OTHER&&this.gamemode==GameMode.NORMAL)
			ai=new LudNormalAI(gd,player);
		else
			ai=new LudSpeedAI(gd,player);
	}
	@Override
	public void getInformation(FrameData frameData) {
		this.frameData = frameData;
		if(frameData.getFrameNumber()>=0){
			if(this.firstflag){this.firstflag=false;firstgetInformation();}//モード判定したり、判定後のフラグ設定をする


		}
		if(ai!=null){
			ai.getInformation(frameData);
		}

	}

	@Override
	public int initialize(GameData arg0, boolean player) {
		System.out.println("test 0603");
		// Initialize the global variables at the start of the round
		inputKey = new Key();
		this.player = player;
		frameData = new FrameData();
		cc = new CommandCenter();
		gd = arg0;

		String tmpcharname=this.gd.getMyName(this.player);
		if(tmpcharname.equals(CHARACTER_ZEN))charname=CharaName.ZEN;
		else if(tmpcharname.equals(CHARACTER_GARNET))charname=CharaName.GARNET;
		else charname=CharaName.OTHER;


		return 0;
	}

	@Override
	public Key input() {
		// The input is set up to the global variable inputKey
		// which is modified in the processing part
		return inputKey;
	}

	public final static int SIMULATE_LIMIT = 60;





	long start;
	long maxtime;
	@Override
	public void processing() {
		// First we check whether we are at the end of the round
		start = System.nanoTime();

		if(!frameData.getEmptyFlag() && frameData.getRemainingTime()>0){
			//System.out.println("frame:"+frameData.getFrameNumber());
			// Simulate the delay and look ahead 2 frames. The simulator class exists already in FightingICE

			if (cc.getskillFlag()) {
				// If there is a previous "command" still in execution, then keep doing it
				inputKey = cc.getSkillKey();
			}
			else {

				inputKey.empty();
				cc.setFrameData(frameData, player);
				cc.skillCancel();
				Action ac=ai.getDoAction();
				String commandstr=(ac.name());



				cc.commandCall(commandstr);
			}
			if(frameData.getFrameNumber()>=0){
				long ttime=System.nanoTime()-start;
				if(ttime>maxtime)maxtime=ttime;
				if(frameData.getFrameNumber()%10==0){
					double timerate=(double)maxtime/(double)(165 * 100000);
					//System.out.println("timerate:\t"+timerate);
					maxtime=0;

				}
			}

		}

	}

}
